/*
 * jets3t : Java Extra-Tasty S3 Toolkit (for Amazon S3 online storage service)
 * This is a java.net project, see https://jets3t.dev.java.net/
 *
 * Copyright 2006 James Murty
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jets3t.service.utils;

import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * A simple container class to store the results generated by
 * {@link FileComparer#buildDiscrepancyLists(Map, Map)}.
 * <p>
 * The following key sets are available:
 * <ul>
 * <li>onlyOnServerKeys: objects with these keys exist only in S3, not on the local computer</li>
 * <li>updatedOnServerKeys: objects with these keys exist in S3 and on the local computer, but the
 *     contents of the objects are different and the S3 version is newer than the local one.</li>
 * <li>onlyOnClientKeys: objects with these keys exist only on the local computer, not in S3</li>
 * <li>updatedOnClientKeys: objects with these keys exist in S3 and on the local computer, but the
 *     contents of the objects are different and the local version is newer than the S3 one.</li>
 * <li>alreadySynchronisedKeys: identical objects with these keys exist in S3 and on the local computer</li>
 * </ul>
 *
 *
 * @author James Murty
 */
public class FileComparerResults {
    public Set onlyOnServerKeys, updatedOnServerKeys, updatedOnClientKeys, onlyOnClientKeys, alreadySynchronisedKeys;

    public FileComparerResults(Set onlyOnServerKeys, Set updatedOnServerKeys,
        Set updatedOnClientKeys, Set onlyOnClientKeys, Set alreadySynchronisedKeys) {
        this.onlyOnServerKeys = onlyOnServerKeys;
        this.updatedOnServerKeys = updatedOnServerKeys;
        this.updatedOnClientKeys = updatedOnClientKeys;
        this.onlyOnClientKeys = onlyOnClientKeys;
        this.alreadySynchronisedKeys = alreadySynchronisedKeys;
    }

    public FileComparerResults() {
        this (new HashSet(), new HashSet(), new HashSet(), new HashSet(), new HashSet());
    }

    public void merge(FileComparerResults resultsToAdd) {
        this.updatedOnServerKeys.addAll(resultsToAdd.updatedOnServerKeys);
        this.updatedOnClientKeys.addAll(resultsToAdd.updatedOnClientKeys);
        this.alreadySynchronisedKeys.addAll(resultsToAdd.alreadySynchronisedKeys);

        this.onlyOnServerKeys.addAll(resultsToAdd.onlyOnServerKeys);

        // Only keys present on S3 and no-where else should remain in server keys list.
        onlyOnServerKeys.removeAll(updatedOnServerKeys);
        onlyOnServerKeys.removeAll(updatedOnClientKeys);
        onlyOnServerKeys.removeAll(alreadySynchronisedKeys);
        onlyOnServerKeys.removeAll(onlyOnClientKeys);

        this.onlyOnClientKeys.addAll(resultsToAdd.onlyOnClientKeys);

        // Only keys present on client and no-where else should remain in client keys list.
        onlyOnClientKeys.removeAll(updatedOnServerKeys);
        onlyOnClientKeys.removeAll(updatedOnClientKeys);
        onlyOnClientKeys.removeAll(alreadySynchronisedKeys);
        onlyOnClientKeys.removeAll(onlyOnServerKeys);
    }

}
